import webbrowser

colfit = Anon(r2_hist_bins=16, to_clean_up={})
def colfit_removeall():
    # fix for v<=1.11.0
    if not (QubX.Charts.fit_ix is None):
        plot = QubX.Charts.plots[QubX.Charts.fit_ix]
        plot.layerset = plot.controlsHidden
    QubX.Charts.remove_all()

def colfit_setup():
    tbl = QubX.Tables.tables[QubX.Tables.book.get_current_page()].table
    if (tbl.label[-7:] == '_FitAll') or (tbl.label[0] == 'z'):
        if not prompt_choices('"%s" might not be the source data table. Continue anyway?' % tbl.label):
            return
    colfit.tbl = tbl
    colfit.tbl_out = QubX.Tables.new_table(name='%s_FitAll' % colfit.tbl.label, initial_row=False)
    colfit.Ncol = len(colfit.tbl.fields) - 2
    if colfit.tbl_out.size != colfit.Ncol:
        colfit.tbl_out.clear()
        if not ('Name' in colfit.tbl_out.fields):
            colfit.tbl_out.add_field('Name', default="", accept=str, format=str, units="")
        for name in colfit.tbl.fields[2:]:
            colfit.tbl_out.append({'Name': name, 'R2': 0, 'SSR': 0})
    colfit_removeall()
    if colfit.Ncol:
        colfit_plot(colfit.tbl.fields[2], True)
    QubX.Charts.request_show()
        
def colfit_plot(name, show_fit):
    # strip trailing zeros
    n = colfit.tbl.size
    while n and not colfit.tbl[n-1, name]:
        n -= 1
    storage = QubX.Tables.new_table(name='z'+name, initial_row=False)
    storage.custom.fitall = colfit.tbl_out
    storage.custom.fitall_index = colfit.tbl.fields.index(name)-2
    colfit.to_clean_up['z'+name] = True
    if storage.size != n:
        storage.clear()
        for i in xrange(n):
            storage.append({name: colfit.tbl[i, name]})
    plot_spec = (storage.label, name, False, 'Index', False)
    QubX.Charts.add_one_plot(*plot_spec)
    plot = QubX.Charts.plots_showing[plot_spec]
    plot.layerset = plot.controls # force fit curve and r2
    if not show_fit:
        plot.layerset = plot.controlsHidden
    return plot

def colfit_cleanup():
    colfit_removeall()
    for name in colfit.to_clean_up:
        t = QubX.Tables.find_table(name)
        if not (t is None):
            QubX.Tables.close_table(t)
    colfit.to_clean_up = {}

def colfit_savecurve():
    if QubX.Charts.fit_ix is None:
        show_message("""Can't find a curve. Please click "Curve fitting" on one of the Charts.""")
    else:
        plot = QubX.Charts.plots[QubX.Charts.fit_ix]
        profile = plot.controls.profile
        colfit.curve = qubx.tree.Node('Curve')
        colfit.curve.appendClone(profile['Name'])
        colfit.curve.appendClone(profile['Eqn'])
        colfit.curve.appendClone(profile['Weight'])
        colfit.curve.appendClone(profile['Params'])
        colfit.curve.appendClone(profile['Strategy'])
        
def colfit_fit(group=None):
    if not colfit.curve:
        show_message("""Please set up a curve and click "save the starting parameters." """)
        return
    colfit_removeall()
    for outrow in colfit.tbl_out:
        if (group is None) or (group == outrow.Group):
            plot = colfit_plot(outrow.Name, True)
            plot.controls.update_profile(plot.controls.profileCat, colfit.curve, curve_only=True)
            plot.controls.robot.sync()
            plot.controls.fit()
            i = outrow.Index
            colfit.tbl_out[i, "R2"] = plot.controls.r2
            colfit.tbl_out[i, "SSR"] = plot.controls.ssr
            for nm, val in izip(plot.controls.param_names, plot.controls.param_vals):
                colfit.tbl_out[i, nm] = val
            plot.layerset = plot.controlsHidden
    colfit_removeall()
    colfit_sort_index(group or 1)

def colfit_refit():
    return colfit_fit(group=1)

def colfit_sort(field, asc=True, cutoff=1):
    colfit_removeall()
    sortix = [i for i in xrange(colfit.tbl_out.size) if (colfit.tbl_out[i, 'Group'] <= cutoff)]
    sortix.sort(lambda i,j: (asc and 1 or -1)*cmp(colfit.tbl_out[i,field], colfit.tbl_out[j,field]))
    for i in sortix:
        colfit_plot(colfit.tbl_out[i, "Name"], False)

def colfit_sort_index(cutoff=1):
    return colfit_sort("Index", True, cutoff)

def colfit_sort_r2d(cutoff=1):
    return colfit_sort("R2", False, cutoff)

def colfit_r2_plot():
    colfit_removeall()
    QubX.Charts.add_one_plot(colfit.tbl_out.label, 'R2', False, 'Index', False)
    QubX.Charts.request_show()

def colfit_r2_hist():
    colfit_removeall()
    QubX.Charts.add_hist_plot(colfit.tbl_out.label, 'R2', False, colfit.r2_hist_bins)
    QubX.Charts.request_show()

def colfit_nbresults():
    for plot in QubX.Charts.plots:
        plot.controls.robot.sync()
        Notebook.send(qubx.notebook.NbItems("", "", plot.nbChart, plot.nbParams, qubx.notebook.NbText("""r^2 = %.3g
""" % plot.controls.r2)))

def colfit_nbsetup():
    Notebook.target_ix = 0
    QubX.Admin.Notebook.request_show()

def colfit_nbshow():
    webbrowser.open('file://%s' % Notebook.target.target.fname)

colfit_workflow = Workflow("ColumnFit",
["""Column Fitting

This window helps automate the process of curve fitting data from columns in a text file.

To start, """, ("Open a table", "QubX.open_table()"), """ of tab- or comma-separated data. The next step, when it's showing in the Tables panel, is to """, ("set up curve fitting", colfit_setup), """.

After you have set up curve fitting (picked a curve and initial parameters), click to """, ("save the starting parameters", colfit_savecurve), """. Then we'll """, ("fit all columns", colfit_fit), """.

""", ("Next: sort and refine", lambda: colfit_workflow.next_page())],
                           ["""Column Fitting

2. Sort and refine

The fitting results are probably onscreen, in a table with the suffix _FitAll, and in the Charts panel. To sort the table, click a column header. To sort the charts, click one of the following: """, ("Sort by r^2 decreasing", colfit_sort_r2d), """, """, ("Sort by Index", colfit_sort_index), """.

We will use the Group column of the FitAll table to direct further processing:
Group 0: successfully fit
Group 1: needs to be re-fit
Group >=2: bad--do not display

You might use a """, ("plot", colfit_r2_plot), """ or """, ("histogram", colfit_r2_hist), """ of r^2 to group the results.

Please review the charts and update the Group column. Click above to re-sort charts. It will hide charts with Group >= 2.

""", ("Next: Re-fit and output", lambda: colfit_workflow.next_page())],
                           ["""Column Fitting

3. Re-fit and output

""", ("Sort by r^2 decreasing", colfit_sort_r2d), """, """, ("Sort by Index", colfit_sort_index), """

If you're going to re-fit, we need different initial parameters or weights; please adjust or re-fit a chart manually, then """, ("click to save curve params", colfit_savecurve), """.

""", ("Re-fit all in Group 1", colfit_refit), """

Finally, arrange the group 0 results for output, """, ("by r^2", lambda: colfit_sort_r2d(0)), """ or """, ("by index", lambda: colfit_sort_index(0)), """, then """,
("write all visible charts to the Notebook", colfit_nbresults), """.

""", ("Set up notebook...", colfit_nbsetup), """
""", ("Show notebook in browser", colfit_nbshow), """

When you're done, """, ("Clean up", colfit_cleanup), """ -- close all charts and temporary tables.
"""])

QubX.Workflows.append_face(colfit_workflow)
